#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import dataclasses
import datetime
import functools
import itertools
import json
import types
import typing
from typing import Any, TypeVar, Union, Callable, Optional

from .annotation import try_extract_type_notes
from .eval_forward_refs import get_evaled_dataclass_fields
from .dataclass_protocol import DataclassProtocol
from .is_instance import get_container_type, is_instance
from .str_to_bool import to_bool


@functools.cache
def get_dataclass_field_name_to_field(
        dataclass_: Any,
        with_init_vars: bool = True, with_class_vars: bool = False
) -> dict[str, dataclasses.Field]:
    try:
        # noinspection PyUnresolvedReferences,PyProtectedMember
        fields = getattr(dataclass_, dataclasses._FIELDS)
    except AttributeError:
        raise TypeError("expected dataclass type")
    # noinspection PyUnresolvedReferences,PyProtectedMember
    allowed_field_types = {dataclasses._FIELD}
    # noinspection PyUnresolvedReferences,PyProtectedMember
    allowed_field_types |= {dataclasses._FIELD_INITVAR} if with_init_vars else set()
    # noinspection PyUnresolvedReferences,PyProtectedMember
    allowed_field_types |= {dataclasses._FIELD_CLASSVAR} if with_class_vars else set()
    # noinspection PyProtectedMember
    result = {name: field for name, field in fields.items() if field._field_type in allowed_field_types}
    return result


class FieldErrors(ValueError):
    def __init__(self, field_to_error: dict[str, Exception]):
        self.field_to_error = field_to_error


class MissingField(ValueError):
    pass


def to_datetime(value: Union[str, int, float]) -> datetime.datetime:
    if isinstance(value, str):
        return datetime.datetime.fromisoformat(value)
    return datetime.datetime.fromtimestamp(value)


def pass_value(value: Any) -> Any:
    return value


@functools.cache
def get_converter(type_: type | types.GenericAlias | types.UnionType) -> Callable[[Any], Any]:
    type_, _ = try_extract_type_notes(type_)
    type_origin = typing.get_origin(type_)
    type_args = typing.get_args(type_)
    if not type_origin and isinstance(type_, type):
        if dataclasses.is_dataclass(type_):
            return functools.partial(dict_to_dataclass, dataclass=type_)
        converter = type_
        if type_ == bool:
            converter = to_bool
        if type_ == typing.Any:
            converter = pass_value
        if type_ == datetime.time:
            converter = datetime.time.fromisoformat
        if type_ == datetime.date:
            converter = datetime.date.fromisoformat
        if type_ == datetime.datetime:
            converter = to_datetime
    elif type_origin is Union or isinstance(type_, types.UnionType):
        def union_converter(value: str):
            if is_instance(value, type_args):
                return value
            for possible_type in type_args:
                try:
                    return get_converter(possible_type)(value)
                except (TypeError, ValueError, AttributeError):
                    pass
            raise ValueError(f"Expected one of: {type_args}, got {type(value)}: {value}")

        return union_converter
    elif type_origin is not None:
        container_type = get_container_type(type_origin)
        if container_type == dict:
            key_converter = get_converter(type_args[0])
            value_converter = get_converter(type_args[1])

            def converter(base_value: Any):
                raw_values = json.loads(base_value) if isinstance(base_value, str) else base_value
                if not isinstance(raw_values, dict):
                    raise ValueError(f'Expected dict, got {type(raw_values).__name__}: {raw_values}')
                parsed_values = [(key_converter(key), value_converter(value)) for key, value in raw_values.items()]
                result = container_type(parsed_values)
                return result
        elif container_type == tuple:
            has_ellipsis = False
            element_converters = []
            for type_ in type_args:
                if type_ is ...:
                    has_ellipsis = True
                else:
                    if has_ellipsis:
                        raise TypeError("Ellipsis (three dots (...)) should be the last"
                                        " in tuple to make it variable-length")
                    element_converters.append(get_converter(type_))
            if has_ellipsis:
                element_converters = itertools.cycle(element_converters)

            def converter(base_value: Any):
                raw_values = json.loads(base_value) if isinstance(base_value, str) else base_value
                if not isinstance(raw_values, (list, tuple, set, frozenset)):
                    raise ValueError(f'Expected iterable, got {type(raw_values).__name__}: {raw_values}')
                if not has_ellipsis and len(raw_values) != len(element_converters):
                    raise ValueError(f'Expected exactly {len(element_converters)} elements, got {len(raw_values)}')
                result = tuple(converter_(element) for element, converter_ in zip(raw_values, element_converters))
                return result
        else:
            element_converter = get_converter(type_args[0])

            def converter(base_value: Any):
                raw_values = json.loads(base_value) if isinstance(base_value, str) else base_value
                if not isinstance(raw_values, (list, tuple, set, frozenset)):
                    raise ValueError(f'Expected iterable, got {type(raw_values).__name__}: {raw_values}')
                parsed_values = [element_converter(element) for element in raw_values]
                result = container_type(parsed_values)
                return result
    elif isinstance(type_, dataclasses.InitVar):
        return get_converter(type_.type)
    else:
        raise TypeError(f'Attempt to cast to unexpected type: {type_} ({type(type_).__name__})')
    return converter


class UnsupportedType(TypeError):
    pass


T = TypeVar('T')
DP = TypeVar('DP', bound=DataclassProtocol)


class NoValue:
    pass


def write_to_dict_or_raise(container: Optional[dict[str, Exception]], key: str, exception: Exception) -> None:
    if container is not None:
        container[key] = exception
    else:
        raise exception


def convert_to_type(field_type: type[T], field_name: str, value: Any,
                    field_name_to_error: dict[str, Exception] = None) -> T:
    field_type, _ = try_extract_type_notes(field_type)
    if dataclasses.is_dataclass(field_type):
        try:
            # noinspection PyTypeChecker
            value = dict_to_dataclass(value, field_type)
        except FieldErrors as e:
            if field_name_to_error is not None:
                for field, error in e.field_to_error.items():
                    write_to_dict_or_raise(field_name_to_error, key=f'{field_name}.{field}', exception=error)
            else:
                raise FieldErrors(
                    {f'{field_name}.{field}': error for field, error in e.field_to_error.items()}) from None
    else:
        try:
            converter = get_converter(field_type)
        except TypeError as e:
            raise UnsupportedType(f'Field "{field_name}": {str(e)}') from e

        try:
            value = converter(value)
        except Exception as e:
            write_to_dict_or_raise(field_name_to_error, field_name, e)
    return value


def convert(name: str, type_: type[T], init: bool, default: Any, default_factory: Callable, value: Any,
            field_name_to_error: dict[str, Exception] = None) -> T:
    if not init:
        return NoValue

    type_, _ = try_extract_type_notes(type_)
    if value != NoValue:
        value = convert_to_type(type_, name, value, field_name_to_error)
    else:
        if default != dataclasses.MISSING:
            return NoValue
        elif default_factory != dataclasses.MISSING:
            return NoValue
        elif dataclasses.is_dataclass(type_):
            try:
                value = dict_to_dataclass({}, type_)
            except FieldErrors as e:
                for field, error in e.field_to_error.items():
                    write_to_dict_or_raise(field_name_to_error, key=f'{name}.{field}', exception=error)
        else:
            write_to_dict_or_raise(field_name_to_error, name, MissingField())
            return NoValue

    return value


def dict_to_dataclass(dict_value: dict[str, Any], dataclass: type[DP]) -> DP:
    dataclass, _ = try_extract_type_notes(dataclass)
    if not dataclasses.is_dataclass(dataclass):
        raise TypeError("expected dataclass type")

    if isinstance(dict_value, dataclass):
        return dict_value

    args = {}
    field_name_to_field = get_dataclass_field_name_to_field(dataclass)
    field_name_to_type = typing.get_type_hints(dataclass) | get_evaled_dataclass_fields(dataclass)
    field_name_to_error = {}
    if hasattr(dict_value, "get"):
        get_value = dict_value.get
    else:
        get_value = lambda name, default: getattr(dict_value, name, default)
    for field_name in field_name_to_field:
        field = field_name_to_field[field_name]
        field_type = field_name_to_type[field_name]
        raw_value = get_value(field_name, NoValue)
        result_value = convert(
            name=field_name, type_=field_type, init=field.init, default=field.default,
            default_factory=field.default_factory, value=raw_value, field_name_to_error=field_name_to_error)
        if result_value is not NoValue:
            args[field_name] = result_value

    if field_name_to_error:
        raise FieldErrors(field_name_to_error)

    return dataclass(**args)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import base64
import datetime
import json
from dataclasses import is_dataclass
from decimal import Decimal
from functools import partial
from typing import Union, TypeAlias, Callable, Protocol, Any

from .dataclass_protocol import DataclassProtocol
from .dict_to_dataclass import get_dataclass_field_name_to_field


JsonLoaded: TypeAlias = Union[list['JsonLoaded'], dict[str, 'JsonLoaded'], str, int, float, None]
Jsonable: TypeAlias = Union[JsonLoaded, list['Jsonable'], tuple['Jsonable', ...], dict[str | int, 'Jsonable']]


def read_json_file_by_path(path: str) -> JsonLoaded:
    with open(path) as file:
        return json.load(file)


class ReprInDumps:
    def __repr_in_dumps__(self):
        return repr(self)


CustomJsonable: TypeAlias = Union[
    Jsonable, list['CustomJsonable'], tuple['CustomJsonable', ...], dict[str | int, 'CustomJsonable'],
    set['CustomJsonable'], frozenset['CustomJsonable'],
    bytes, ReprInDumps, type, DataclassProtocol, datetime.datetime, datetime.date, datetime.time, Decimal, Exception
]


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ReprInDumps):
            return o.__repr_in_dumps__()
        if isinstance(o, type):
            return {'_t': 'PY::class', 'key': o.__qualname__}
        if is_dataclass(o):
            return {
                key: getattr(o, key)
                for key, field in get_dataclass_field_name_to_field(type(o), with_init_vars=False).items()
                if field.repr
            }
        if isinstance(o, datetime.datetime):
            if o.tzinfo is None or o.tzinfo.utcoffset(o) is None:  # PROHIBIT naive datetime serialisation
                raise TypeError("TypeError: datetime.datetime WITHOUT tzinfo is not JSON serializable")
        if isinstance(o, datetime.date | datetime.time):  # date or time or datetime
            return o.isoformat()
        if isinstance(o, set | frozenset):
            return tuple(o)
        if isinstance(o, bytes):
            return f'data:application/octet-stream;base64,{base64.b64encode(o)}'
        if isinstance(o, Decimal):
            return str(o)
        if isinstance(o, Exception):
            return {'_t': 'PY::Exception', 'key': type(o).__qualname__, 'args': o.args}
        return super().default(o)


class JsonDumper(Protocol):
    @staticmethod
    def __call__(
            obj: CustomJsonable, *, skipkeys: bool = False, ensure_ascii: bool = True, check_circular: bool = True,
            allow_nan: bool = True, indent: int | None = None, separators: tuple[str, str] | None = None,
            default: Callable[[Any], Jsonable] = None, sort_keys: bool = False, **kw
    ) -> str:
        ...


custom_dumps: JsonDumper = partial(json.dumps, cls=EnhancedJSONEncoder)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import logging
import pathlib
import tomllib
from typing import Dict

logger = logging.getLogger(__file__)


def read_toml_file_by_path(path: str) -> Dict[str, Dict[str, str]]:
    logger.info(f"Attempt to read toml file by {path=}")
    text = pathlib.Path(path).read_text()
    return tomllib.loads(text)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
from init_helpers.annotation import note, try_extract_type_notes
from init_helpers.any_type import TypingGenericAlias, AnyType
from init_helpers.conditional_insert import conditional_set, conditional_append, conditional_add
from init_helpers.conditional_insert import set_if, append_if, add_if
from init_helpers.custom_json import ReprInDumps, custom_dumps, read_json_file_by_path, Jsonable
from init_helpers.dataclass_protocol import DataclassProtocol
from init_helpers.dict_to_dataclass import dict_to_dataclass, get_converter, write_to_dict_or_raise, convert
from init_helpers.eval_forward_refs import eval_forward_refs_in_local_dataclasses, get_evaled_dataclass_fields
from init_helpers.get_instance_id import InstanceIdConfig, get_instance_id
from init_helpers.get_non_generic_args import get_non_generic_args
from init_helpers.ini_reader import read_ini_file_by_path
from init_helpers.init_random import init_random
from init_helpers.init_logs import LogsConfig, init_logs
from init_helpers.is_instance import is_instance, is_collection, is_tuple, is_mapping, get_container_type
from init_helpers.is_none import is_none, is_not_none
from init_helpers.parse_args_as_dataclass import parse_args_as_dataclass
from init_helpers.pickable_dict import PickableDict
from init_helpers.parse_args import parse_args, Arg
from init_helpers.exceptions import repr_exception
from init_helpers.raise_if import raise_if
from init_helpers.read_file_text import read_file_text
from init_helpers.str_to_bool import str_to_bool
from init_helpers.structure_hint import StructureHint, FieldHint, hint_to_ini, dataclass_to_hint
from init_helpers.structure_hint import get_dataclass_field_name_to_field, get_name_from_type
from init_helpers.to_list import to_list
from init_helpers.toml_reader import read_toml_file_by_path
from init_helpers.try_cast import try_cast
from init_helpers.unflatten_dict import unflatten_dict
from init_helpers.update_dict_recursive import update_dict_recursive
from init_helpers.url_to_snake_case import url_to_snake_case



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import os
import random
import time

from init_helpers.dict_to_dataclass import NoValue


def init_random(
        seed: None | int | float | str | bytes | bytearray | NoValue
) -> None | int | float | str | bytes | bytearray:
    seed = os.getpid() + time.time() if seed is NoValue else seed
    random.seed(seed)
    return seed



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import types
import typing

TypingGenericAlias: typing.TypeAlias = type(typing.List[str])
AnyType: typing.TypeAlias = type | types.UnionType | types.GenericAlias | TypingGenericAlias  # OR typing.Annotated



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import configparser
import logging
import os
import pathlib

from .read_file_text import read_file_text

logger = logging.getLogger(__file__)


class ParseError(Exception):
    pass


class IniParseError(ParseError):
    pass


def parse_ini_content(ini_content: str) -> dict[str, dict[str, str]]:
    config = configparser.RawConfigParser()
    try:
        config.read_string(ini_content)
    except configparser.Error as e:
        raise IniParseError(f"Failed to parse {ini_content=} as dict: {e=}") from None
    return {name: {key: config[name][key] for key in config[name]} for name in config.sections()}


def read_ini_file_by_path(
        path: str | pathlib.Path | os.PathLike, encoding: str | None = None
) -> dict[str, dict[str, str]]:
    logger.info(f"Attempt to read ini file by {path=}")
    return parse_ini_content(read_file_text(path=path, encoding=encoding))



import collections.abc
import itertools
import types
from typing import Any, Iterable, Union, Sequence, get_args, get_origin

from init_helpers.to_list import to_list


def get_container_type(type_origin) -> type:
    if not getattr(type_origin, "__abstractmethods__", None):
        return type_origin
    container_type = ORIGIN_TO_CONTAINER_TYPE.get(type_origin)
    if container_type is None:
        raise TypeError(f'Attempt to get_container_type from unexpected origin: {type_origin}')

    return container_type


def is_tuple(value: Any, tuple_args: Iterable[Union[type, types.GenericAlias]]) -> bool:
    if not isinstance(value, tuple):
        return False

    has_ellipsis = False
    type_checks = []
    for inner_type in tuple_args:
        if inner_type is ...:
            has_ellipsis = True
        else:
            if has_ellipsis:
                raise TypeError("Ellipsis (three dots (...)) should be the last"
                                " in tuple to make it variable-length")
            type_checks.append(inner_type)
    if has_ellipsis:
        type_checks = itertools.cycle(type_checks)
    if not has_ellipsis and type_checks:
        if len(type_checks) != len(value):
            return False

    for val, inner_type in zip(value, type_checks):
        if not is_instance(val, [inner_type]):
            return False

    return True


def is_mapping(value: Any, mapping_args: Sequence[Union[type, types.GenericAlias]]) -> bool:
    if not isinstance(value, collections.abc.Mapping):
        return False

    if not mapping_args:
        return True

    if len(mapping_args) != 2:
        raise TypeError(f"Mapping requires exactly 2 type args, got: {mapping_args}")

    key_type_list = [mapping_args[0]]
    value_type_list = [mapping_args[1]]

    for key, val in value.items():
        if not is_instance(key, key_type_list) or not is_instance(val, value_type_list):
            return False
    return True


def is_collection(value: Any, collection_args: Sequence[Union[type, types.GenericAlias]]) -> bool:
    if not isinstance(value, collections.abc.Collection):
        return False

    if not collection_args:
        return True

    if len(collection_args) != 1:
        raise TypeError(f"Collection requires exactly 2 type args, got: {collection_args}")

    collection_type_list = [collection_args[0]]

    for val in value:
        if not is_instance(val, collection_type_list):
            return False
    return True


def is_instance(
        val: Any,
        possible_types: type | types.GenericAlias | types.UnionType | Iterable[
                        type | types.GenericAlias | types.UnionType]
) -> bool:
    for type_ in to_list(possible_types):
        if type_ == Any:
            return True

        type_origin = get_origin(type_)
        type_args = get_args(type_)
        if not type_origin and isinstance(type_, type):  # basic check
            if isinstance(val, type_):
                return True
        elif type_origin is Union or isinstance(type_, types.UnionType):
            if is_instance(val, type_args):
                return True
        else:
            container_type = get_container_type(type_origin)
            if container_type is None or not isinstance(val, container_type):
                continue

            if issubclass(container_type, collections.abc.Mapping):
                if is_mapping(val, type_args):
                    return True

            elif container_type is tuple:  # tuple has specific typing behaviour
                if is_tuple(val, type_args):
                    return True

            elif issubclass(container_type, collections.abc.Collection):
                if is_collection(val, type_args):
                    return True
    return False


ORIGIN_TO_CONTAINER_TYPE: dict[collections.abc.Container, type] = {
    collections.abc.Mapping: dict,
    collections.abc.Collection: list,
    collections.abc.Set: set,
}



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>

def update_dict_recursive(target_dict: dict, source_dict: dict) -> None:
    for key, value in source_dict.items():
        if isinstance(value, dict) and isinstance(target_value := target_dict.get(key), dict):
            update_dict_recursive(target_value, value)
        else:
            target_dict[key] = value




#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import argparse
from typing import TypeVar, Iterable

from .ini_reader import read_ini_file_by_path
from .dataclass_protocol import DataclassProtocol
from .dict_to_dataclass import dict_to_dataclass
from .toml_reader import read_toml_file_by_path
from .update_dict_recursive import update_dict_recursive
from .unflatten_dict import unflatten_dict

DP = TypeVar('DP', bound=DataclassProtocol)


def parse_args_as_dataclass(dataclass_type: type[DP], flag_name: str = '-v', args: Iterable[str] | None = None) -> DP:
    parser = argparse.ArgumentParser()
    parser.add_argument('config_file_path', nargs='?', default=None)
    overload_arg = parser.add_argument(flag_name, action='append')
    parsed = parser.parse_args(args)
    config = read_config(parsed.config_file_path) if parsed.config_file_path else {}
    raw_overload = getattr(parsed, overload_arg.dest) or []
    overload: dict[str, str | dict[str, str | dict]] = unflatten_dict(dict([val.split('=') for val in raw_overload]))
    update_dict_recursive(config, overload)
    return dict_to_dataclass(config, dataclass_type)


def read_config(config_file_path: str) -> dict:
    extension = config_file_path.rsplit('.', 1)[-1]
    if extension == 'toml':
        return read_toml_file_by_path(config_file_path)
    if extension == 'ini':
        return read_ini_file_by_path(config_file_path)
    raise TypeError(f'Unknown {extension=}')



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import logging
import os
from logging.handlers import RotatingFileHandler

import sys
from dataclasses import dataclass
from typing import Optional


@dataclass
class LogsConfig:
    level: str = "DEBUG"
    filename: str = None
    max_file_size_megabytes: int = 10
    max_files_amount: int = 10


class CustomFormatter(logging.Formatter):
    def __init__(self, msg=None, *args, **kwargs):
        if msg is None:
            msg = '<%(asctime)s/%(name)s:%(lineno)s/%(levelname)s>-  %(message)s'
        logging.Formatter.__init__(self, msg, *args, **kwargs)

    def format(self, record: logging.LogRecord):
        record.name = f'{os.getpid()}/{record.name.rsplit(".")[0].rsplit("/")[-1]}'
        return logging.Formatter.format(self, record)


def init_logs(settings: Optional[LogsConfig] = None, logging_formatter: Optional[logging.Formatter] = None) -> None:
    if settings is None:
        settings = LogsConfig()

    root = logging.getLogger()
    root.setLevel(settings.level)

    if settings.filename is None:
        ch = logging.StreamHandler(sys.stdout)
    else:
        max_file_bytes = settings.max_file_size_megabytes * 1024 * 1024
        ch = RotatingFileHandler(settings.filename, maxBytes=max_file_bytes, backupCount=settings.max_files_amount)
    ch.setLevel(logging.getLevelName(settings.level))

    if logging_formatter is None:
        ch.setFormatter(CustomFormatter(datefmt='%Y-%m-%d %H:%M:%S'))
    else:
        ch.setFormatter(logging_formatter)
    root.addHandler(ch)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
from typing import TypeVar, MutableMapping, Callable, Any

from init_helpers.is_none import is_not_none


T = TypeVar('T')
Condition = Callable[[T], bool]


def conditional_set(mapping: MutableMapping[str, Any], *, _condition: Condition = is_not_none, **kwargs: T) -> None:
    for key, value in kwargs.items():
        if _condition(value):
            mapping[key] = value


def conditional_append(list_: list, *args: T, _condition: Condition = is_not_none) -> None:
    for value in args:
        if _condition(value):
            list_.append(value)


def conditional_add(set_: set, *args: T, _condition: Condition = is_not_none) -> None:
    for value in args:
        if _condition(value):
            set_.add(value)


def set_if(condition: bool, mapping: MutableMapping[str, Any], **kwargs: Any) -> None:
    if condition:
        for key, value in kwargs.items():
            mapping[key] = value


def append_if(condition: bool, list_: list[Any], *args: Any) -> None:
    if condition:
        for value in args:
            list_.append(value)


def add_if(condition: bool, set_: set[Any], *args: Any) -> None:
    if condition:
        for value in args:
            set_.add(value)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import collections.abc
import dataclasses
import types
import typing
from dataclasses import dataclass
from types import UnionType, NoneType
from typing import Union, Any

from .dict_to_dataclass import get_dataclass_field_name_to_field


@dataclass(slots=True, frozen=True)
class FieldHint:
    name: str
    type: type | types.UnionType | dataclasses.InitVar  # | typing.Union
    default: Any = dataclasses.field(default_factory=lambda: dataclasses.MISSING)

    def __eq__(self, other: 'FieldHint') -> bool:
        # InitVar[str] != InitVar[str], so make distinct type check
        if (not isinstance(other, FieldHint) or self.name != other.name or
                type(self.type) != type(self.type) or self.default != other.default):
            return False
        if isinstance(self.type, dataclasses.InitVar):
            return self.type.type == other.type.type
        return self.type == other.type

    def __repr__(self) -> str:
        parts = []
        parts += [f'name={self.name!r}']
        parts += [f'type={self.type.__name__ if isinstance(self.type, type) else self.type}']
        parts += [] if self.default is dataclasses.MISSING else [f'default={self.default!r}']
        return f'{type(self).__name__}({", ".join(parts)})'

    __str__ = __repr__


@dataclass(slots=True, frozen=True)
class StructureHint:
    fields: collections.abc.Sequence['FieldHint | StructureHint']
    name: str | None = None

    def __post_init__(self):
        object.__setattr__(self, 'fields', tuple(self.fields))

    @staticmethod
    def from_dataclass(dataclass_type: type) -> 'StructureHint':
        return dataclass_to_hint(dataclass_type)

    def __repr__(self) -> str:
        parts = []
        parts += [] if self.name is None else [f'name={self.name!r}']
        parts += [f'fields={list(self.fields)}']
        return f'{type(self).__name__}({", ".join(parts)})'

    def to_ini(self) -> str:
        return hint_to_ini(self)


def get_name_from_type(type_: type) -> str:
    type_origin = typing.get_origin(type_)
    type_args = typing.get_args(type_)
    if type_origin is not None and type_args is not None:
        if type_origin in (Union, UnionType):
            return ' | '.join(list(get_name_from_type(sub_type) for sub_type in type_args))
        result = get_name_from_type(type_origin)
        if type_args:
            result += '[' + (', '.join(list(get_name_from_type(sub_type) for sub_type in type_args))) + ']'
        return result
    elif isinstance(type_, dataclasses.InitVar):
        return get_name_from_type(type_.type)
    elif type_ in (None, NoneType):
        return "None"
    elif type_ is Ellipsis:
        return "..."
    return type_.__name__


def hint_to_ini(hint: StructureHint, depth: int = 0) -> str:
    section_name_to_lines: dict[str, list[str]] = {}
    section: list[str] | None = None
    for f in hint.fields:
        # if res:
        #     res += "\n"
        if isinstance(f, FieldHint):
            section = section or section_name_to_lines.setdefault('', [])
            default = "" if f.default is dataclasses.MISSING else f", default = {f.default}"
            section.append(f'; {f.name}: {get_name_from_type(f.type)}{default}')
            section.append(f'{f.name} = ...')
        elif isinstance(f, StructureHint):
            if depth:
                raise ValueError(f"INI does not support more than one nesting level, got at least two: {f.name}")
            section_name_to_lines[f.name] = [hint_to_ini(f, depth+1)]
        else:
            raise TypeError(f"unexpected type: {type(f)}")

    blocks = []
    for section_name, lines in section_name_to_lines.items():
        result = ''
        if section_name:
            result += f'[{section_name}]\n'
        result += '\n'.join(lines)
        blocks.append(result)
    return '\n\n'.join(blocks)


def dataclass_to_hint(dataclass_type: type) -> StructureHint:
    if not dataclasses.is_dataclass(dataclass_type):
        raise TypeError("expected dataclass type")

    args = []
    dataclass_fields = get_dataclass_field_name_to_field(dataclass_type)
    for field_name in dataclass_fields:
        field = dataclass_fields[field_name]
        if not field.init:
            continue

        default = dataclasses.MISSING
        if field.default != dataclasses.MISSING:
            default = field.default
        elif field.default_factory != dataclasses.MISSING:
            default = field.default_factory

        field_type = field.type
        if dataclasses.is_dataclass(field_type):
            hint = dataclass_to_hint(field_type)
            hint = dataclasses.replace(hint, name=field_name)
        else:
            hint = FieldHint(field_name, field_type, default)

        args.append(hint)
    return StructureHint(args)
